/*-- List.h ---------------------------------------------------------------
 
 This header file defines the data type List for processing lists.
 Basic operations are:
 Constructor
 empty:    Check if list is empty
 insert:   Insert an item
 erase:    Remove an item
 
 << :  output operator
 -------------------------------------------------------------------------*/

#include <iostream>

#ifndef LIST_H
#define LIST_H

using namespace std;

const int CAPACITY = 1024;
template <class ElementType>
class List
{
public:
    /******** Function Members ********/
    /***** Class constructor *****/
    List();
    /*----------------------------------------------------------------------
     Construct a List object.
     
     Precondition:  None
     Postcondition: An empty List object has been constructed;
     mySize is 0.
     -----------------------------------------------------------------------*/
    
    /***** empty operation *****/
    bool empty() const;
    /*----------------------------------------------------------------------
     Check if a list is empty.
     
     Precondition:  None
     Postcondition: true is returned if the list is empty,
     false if not.
     -----------------------------------------------------------------------*/
    
    /***** insert and erase *****/
    void insert(ElementType item, int pos);
    /*----------------------------------------------------------------------
     Insert a value into the list at a given position.
     
     Precondition:  item is the value to be inserted; there is room in
     the array (mySize < CAPACITY); and the position satisfies
     0 <= pos <= mySize.
     Postcondition: item has been inserted into the list at the position
     determined by pos (provided there is room and pos is a legal
     position).
     -----------------------------------------------------------------------*/
    
    void erase(int pos);
    /*----------------------------------------------------------------------
     Remove a value from the list at a given position.
     
     Precondition:  The list is not empty and the position satisfies
     0 <= pos < mySize.
     Postcondition: element at the position determined by pos has been
     removed (provided pos is a legal position).
     ----------------------------------------------------------------------*/
    
    /***** output *****/
    void display(ostream & out) const;
    /*----------------------------------------------------------------------
     Display a list.
     
     Precondition:  out is a reference parameter
     Postcondition: The list represented by this List object has been
     inserted into ostream out.
     -----------------------------------------------------------------------*/
    
private:
    /******** Data Members ********/
    int mySize;                     // current size of list stored in myArray
    ElementType myArray[CAPACITY];  // array to store list elements
    
}; //--- end of List class

//------ Prototype of output operator
template <class ElementType>
ostream & operator<< (ostream & out, const List<ElementType> & aList);
/*-- Implementation of List member functions.
 
 -------------------------------------------------------------------------*/


//--- Definition of class constructor

template <class ElementType>
List<ElementType>::List()
: mySize(0)
{
}

//--- Definition of empty()
template <class ElementType>
bool List<ElementType>::empty() const
{
    return mySize == 0;
}

template <class ElementType>
void List<ElementType>::display(ostream & out) const
{
    for (int i = 0; i < mySize; i++)
        out << myArray[i] << "  ";
}


//--- Definition of output operator
template <class ElementType>
ostream & operator<< (ostream & out, const List<ElementType> & aList)
{
    aList.display(out);
    return out;
}

//--- Definition of insert()
template <class ElementType>
void List<ElementType>::insert(ElementType item, int pos)
{
    if (mySize == CAPACITY)
    {
        cerr << "*** No space for list element -- terminating "
        "execution ***\n";
        exit(1);
    }
    if (pos < 0 || pos > mySize)
    {
        cerr << "*** Illegal location to insert -- " << pos
        << ".  List unchanged. ***\n";
        return;
    }
    
    // First shift array elements right to make room for item
    
    for(int i = mySize; i > pos; i--)
        myArray[i] = myArray[i - 1];
    
    // Now insert item at position pos and increase list size
    myArray[pos] = item;
    mySize++;
}

// Definition of erase()
template <class ElementType>
void List<ElementType>::erase(int pos)
{
    if (mySize == 0)
    {
        cerr << "*** List is empty ***\n";
        return;
    }
    if (pos < 0 || pos >= mySize)
    {
        cerr << "Illegal location to delete -- " << pos
        << ".  List unchanged. ***\n";
        return;
    }
    
    // Shift array elements left to close the gap
    for(int i = pos; i < mySize; i++)
        myArray[i] = myArray[i + 1];
    
    // Decrease list size
    mySize--;
}

#endif

