/*-- List.h ---------------------------------------------------------------
 
 This header file defines the data type List for processing lists.
 Basic operations are:
 Constructor
 empty:    Check if list is empty
 insert:   Insert an item
 erase:    Remove an item
 display:  Output the list
 <<, >> :  input and output operators
 -------------------------------------------------------------------------*/

#include <iostream>
#include <cassert>

#ifndef LIST_H
#define LIST_H

using namespace std;

template <class ElementType>
class List
{
public:
    /******** Function Members ********/
    /***** Class constructor *****/
    List(int maxSize = 1024);
    /*----------------------------------------------------------------------
     Construct a List object.
     
     Precondition:  maxSize is a positive integer with default value 1024.
     Postcondition: An empty List object is constructed; myCapacity ==
     maxSize (default value 1024); myArrayPtr points to a run-time
     array with myCapacity as its capacity; and mySize is 0.
     -----------------------------------------------------------------------*/
    /***** Class destructor *****/
    ~List();
    /*----------------------------------------------------------------------
     Destroys a List object.
     
     Precondition:  The life of a List object is over.
     Postcondition: The memory dynamically allocated by the constructor
     for the array pointed to by myArrayPtr has been returned to
     the heap.
     ------------------------------------------------------------------*/
    /***** Copy constructor *****/
    List(const List<ElementType> & origList);
    /*----------------------------------------------------------------------
     Construct a copy of a List object.
     
     Precondition:  A copy of origList is needed; origList is a const
     reference parameter.
     Postcondition: A copy of origList has been constructed.
     -----------------------------------------------------------------------*/
    /***** Assignment operator *****/
    List & operator=(const List<ElementType> & origList);
    /*----------------------------------------------------------------------
     Assign a copy of a List object to the current object.
     
     Precondition: none
     Postcondition: A copy of origList has been assigned to this
     object.  A reference to this list is returned.
     -----------------------------------------------------------------------*/
    
    /***** empty operation *****/
    bool empty() const;
    /*----------------------------------------------------------------------
     Check if a list is empty.
     
     Precondition:  None
     Postcondition: true is returned if the list is empty,
     false if not.
     -----------------------------------------------------------------------*/
    
    /***** insert and erase *****/
    void insert(ElementType item, int pos);
    /*----------------------------------------------------------------------
     Insert a value into the list at a given position.
     
     Precondition:  item is the value to be inserted; there is room in
     the array (mySize < CAPACITY); and the position satisfies
     0 <= pos <= mySize.
     Postcondition: item has been inserted into the list at the position
     determined by pos (provided there is room and pos is a legal
     position).
     -----------------------------------------------------------------------*/
    
    void erase(int pos);
    /*----------------------------------------------------------------------
     Remove a value from the list at a given position.
     
     Precondition:  The list is not empty and the position satisfies
     0 <= pos < mySize.
     Postcondition: element at the position determined by pos has been
     removed (provided pos is a legal position).
     ----------------------------------------------------------------------*/
    
    /***** output *****/
    void display(ostream & out) const;
    /*----------------------------------------------------------------------
     Display a list.
     
     Precondition:  out is a reference parameter
     Postcondition: The list represented by this List object has been
     inserted
    

    
private:
    /******** Data Members ********/
    int mySize;                // current size of list stored in array
    int myCapacity;            // capacity of array
    ElementType * myArrayPtr;  // pointer to dynamically-allocated arraylements
    
}; //--- end of List class

template< class ElementType>
ostream & operator<< (ostream & out, const List<ElementType> & aList);

/*-- Implementation of List member functions.
 
 -------------------------------------------------------------------------*/


//--- Definition of class constructor

template <class ElementType>
List<ElementType>::List(int maxSize)
: mySize(0), myCapacity(maxSize)
{
    myArrayPtr = new(nothrow) ElementType[maxSize];
    assert(myArrayPtr != 0);
}

//--- Definition of class destructor
template <class ElementType>
List<ElementType>::~List()
{
    delete [] myArrayPtr;
}

//--- Definition of the copy constructor
template <class ElementType>
List<ElementType>::List(const List<ElementType> & origList)
: mySize(origList.mySize), myCapacity(origList.myCapacity)

{
    //--- Get new array for copy
    myArrayPtr = new(nothrow) ElementType[myCapacity];
    
    if (myArrayPtr != 0)                 // check if memory available
        //--- Copy origList's array into this new array
        for(int i = 0; i < myCapacity; i++)
            myArrayPtr[i] = origList.myArrayPtr[i];
    else
    {
        cerr << "*Inadequate memory to allocate stack ***\n";
        exit(1);
    }
}

//--- Definition of the assignment operator
template <class ElementType>
List<ElementType> & List<ElementType>::operator=(const List<ElementType> & origList)
{
    if (this != &origList)    // check for list = list
    {
        mySize = origList.mySize;
        
        //-- Allocate a new array if necessary
        if (myCapacity != origList.myCapacity)
        {
            delete[] myArrayPtr;
            myCapacity = origList.myCapacity;
            myArrayPtr = new(nothrow) ElementType[myCapacity];
            if (myArrayPtr == 0)   // check if memory available
            {
                cerr << "*Inadequate memory to allocate stack ***\n";
                exit(1);
            }
        }
        //--- Copy origList's array into this new array
        for(int i = 0; i < myCapacity; i++)
            myArrayPtr[i] = origList.myArrayPtr[i];
    }
    return *this;
}

//--- Definition of empty()
template <class ElementType>
bool List<ElementType>::empty() const
{
    return mySize == 0;
}


template <class ElementType>
void List<ElementType>::display(ostream & out) const
{
    for (int i = 0; i < mySize; i++)
        out << myArrayPtr[i] << "  ";
}


//--- Definition of output operator
template <class ElementType>
ostream & operator<< (ostream & out, const List<ElementType> & aList)
{
    aList.display(out);
    return out;
}

//--- Definition of insert()
template <class ElementType>
void List<ElementType>::insert(ElementType item, int pos)
{
    if (mySize == myCapacity)
    {
        cerr << "*** No space for list element -- terminating "
        "execution ***\n";
        exit(1);
    }
    if (pos < 0 || pos > mySize)
    {
        cerr << "*** Illegal location to insert -- " << pos
        << ".  List unchanged. ***\n";
        return;
    }
    
    // First shift array elements right to make room for item
    
    for(int i = mySize; i > pos; i--)
        myArrayPtr[i] = myArrayPtr[i - 1];
    
    // Now insert item at position pos and increase list size
    myArrayPtr[pos] = item;
    mySize++;
}

// Definition of erase()
template <class ElementType>
void List<ElementType>::erase(int pos)
{
    if (mySize == 0)
    {
        cerr << "*** List is empty ***\n";
        return;
    }
    if (pos < 0 || pos >= mySize)
    {
        cerr << "Illegal location to delete -- " << pos
        << ".  List unchanged. ***\n";
        return;
    }
    
    // Shift array elements left to close the gap
    for(int i = pos; i < mySize; i++)
        myArrayPtr[i] = myArrayPtr[i + 1];
    
    // Decrease list size
    mySize--;
}

#endif

