/*-- ListLinked.h ---------------------------------------------------------------
 
 This header file defines the data type List for processing lists.
 Basic operations are:
 Constructor, Copy, Assignment, Destructor
 empty:    Check if list is empty
 insert:   Insert an item
 erase:    Remove an item
 display:  Output the list
 <<, >> :  input and output operators
 -------------------------------------------------------------------------*/

#include <iostream>
#include <cassert>
#ifndef ListLinked_h
#define ListLinked_h

using namespace std;

template <class ElementType>
class List
{
public:
    
    /******** Function Members ********/
    /***** Class constructor *****/
    List();
    /*----------------------------------------------------------------------
     Construct a List object.
     
     Precondition:  none
     Postcondition: An empty List object is constructed;      -----------------------------------------------------------------------*/
    /***** Class destructor *****/
    ~List();
    /*----------------------------------------------------------------------
     Destroys a List object.
     
     Precondition:  The life of a List object is over.
     Postcondition: The memory dynamically allocated by the constructor has been returned to
     the heap.
     ------------------------------------------------------------------*/
    /***** Copy constructor *****/
    List(const List<ElementType> & origList);
    /*----------------------------------------------------------------------
     Construct a copy of a List object.
     
     Precondition:  A copy of origList is needed; origList is a const
     reference parameter.
     Postcondition: A copy of origList has been constructed.
     -----------------------------------------------------------------------*/
    /***** Assignment operator *****/
    List & operator=(const List<ElementType> & origList);
    /*----------------------------------------------------------------------
     Assign a copy of a List object to the current object.
     
     Precondition: none
     Postcondition: A copy of origList has been assigned to this
     object.  A reference to this list is returned.
     -----------------------------------------------------------------------*/
    
    /***** empty operation *****/
    bool empty() const;
    /*----------------------------------------------------------------------
     Check if a list is empty.
     
     Precondition:  None
     Postcondition: true is returned if the list is empty,
     false if not.
     -----------------------------------------------------------------------*/
    
    /***** insert and erase *****/
    void insert(ElementType item, int pos);
    /*----------------------------------------------------------------------
     Insert a value into the list at a given position.
     
     Precondition:  item is the value to be inserted; and the position satisfies
     0 <= pos <= mySize.
     Postcondition: item has been inserted into the list at the position
     determined by pos (provided pos is a legal
     position).
     -----------------------------------------------------------------------*/
    
    void erase(int pos);
    /*----------------------------------------------------------------------
     Remove a value from the list at a given position.
     
     Precondition:  The list is not empty and the position satisfies
     0 <= pos < mySize.
     Postcondition: element at the position determined by pos has been
     removed (provided pos is a legal position).
     ----------------------------------------------------------------------*/
    
    /***** output *****/
    void display(ostream & out) const;
    /*----------------------------------------------------------------------
     Display a list.
     
     Precondition:  out is a reference parameter
     Postcondition: The list represented by this List object has been
     inserted*/
     
     private:
    
     class ListNode {
       public:
         
         //Node constructor
        ListNode(const ElementType& nodeData, ListNode* nextPtr);
        
         /******** Node Data Members ********/
        ElementType dataItem;
        ListNode* next;
    };

     /******** List Data Members ********/
    int mySize;                // current size of list
    ListNode * head;  // pointer to head of list
}; //--- end of List class

template< class ElementType>
ostream & operator<< (ostream & out, const List<ElementType> & aList);


//--- Definition of Node class constructor
template <typename ElementType>
List<ElementType>::ListNode::ListNode(const ElementType& nodeData,ListNode* nextPtr)
// Creates a list node containing item elem and next pointer
// nextPtr.

: dataItem(nodeData), next(nextPtr)
{
}

/*-- Implementation of List member functions.
 
 -------------------------------------------------------------------------*/
//--- Definition of List class constructor

template <class ElementType>
List<ElementType>::List()
: mySize(0), head(nullptr)
{

}

//--- Definition of class destructor
template <class ElementType>
List<ElementType>::~List()
{
    ListNode *cursor = head;
    while ( cursor != nullptr )
    {
        head = head->next;
        delete cursor;
        cursor = head;
    }
}

//--- Definition of the copy constructor
template <class ElementType>
List<ElementType>::List(const List<ElementType> & origList)
: mySize(origList.mySize), head(nullptr)

{
    //a bit of cheating
    operator=(origList);
}

//--- Definition of the assignment operator
template <class ElementType>
List<ElementType> & List<ElementType>::operator=(const List<ElementType> & origList)
{
    if( this != &origList ) {
        
        // clear this list - could use iterated erase
        ListNode *cursor = head;
        while ( cursor != nullptr )
        {
            head = head->next;
            delete cursor;
            cursor = head;
        }
        mySize = 0;
        head = nullptr;
        //Point to head of List to be copied
        ListNode *otherPtr = origList.head;
        
        int pos = 0;
        while( otherPtr != nullptr ) {
            insert(otherPtr->dataItem, pos);
            pos++;
            otherPtr = otherPtr->next;
        }
        
    }
    
    return *this;
}

//--- Definition of empty()
template <class ElementType>
bool List<ElementType>::empty() const
{
    return mySize == 0;
}


template <class ElementType>
void List<ElementType>::display(ostream & out) const
{
    if ( empty() )
    {
        out << "Empty list" << endl;
    }
    else
    {
        for (ListNode* temp = head; temp != nullptr; temp = temp->next) {
            
            // Assumes that dataItem can be printed via << because
            // is is either primitive or operator<< is overloaded.
            out << temp->dataItem;
            out << " ";
        }
        out << endl;
    }
}


//--- Definition of output operator
template <class ElementType>
ostream & operator<< (ostream & out, const List<ElementType> & aList)
{
    aList.display(out);
    return out;
}

//--- Definition of insert()
template <class ElementType>
void List<ElementType>::insert(ElementType item, int pos)
{
    
    if (pos < 0 || pos > mySize)
    {
        cerr << "*** Illegal location to insert -- " << pos
        << ".  List unchanged. ***\n";
        return;
    }
    
    ListNode *temp = new ListNode(item, nullptr);
    
    // CK - Insert at front
    if (pos == 0){
        //check if empty or not
        if(head == nullptr)
            head = temp;
        else {
            temp->next = head;
            head = temp;
        }
    }
    else {
        //move to correct position
        ListNode *cursor = head;
        for(int i = 1; i < pos; i++) {
            cursor = cursor->next;
        }
        temp->next = cursor->next;
        cursor->next = temp;
    }
    
    mySize++;
}

// Definition of erase()
template <class ElementType>
void List<ElementType>::erase(int pos)
{
    
    
    if (pos < 0 || pos > mySize-1)
    {
        cerr << "*** Illegal location to erase -- " << pos
        << ".  List unchanged. ***\n";
        return;
    }
    
    if(empty())
    {
        cerr << "*** Can't erase from an empty list "
        << ".  List unchanged. ***\n";
        return;
    }
    
    if(pos == 0 )
    {
        ListNode *temp = head;
        head = head->next;
        delete temp;
        mySize--;
        
    }
    else
    {
        //move to correct position
        ListNode *cursor = head;
        for(int i = 1; i < pos; i++)
        {
            cursor = cursor->next;
        }
        
        ListNode *temp = cursor->next;
        cursor->next = temp->next;
        delete temp;
        mySize--;
    }
}

#endif

